r"""
    Dataloading utilities for evaluator
"""
from typing import Tuple, Optional, Union
from pathlib import Path
import numpy as np
import pandas as pd
from pynwb import NWBHDF5IO

from falcon_challenge.config import FalconTask

# Load nwb file
def bin_units(
        units: pd.DataFrame,
        bin_size_s: float = 0.02,
        bin_timestamps: Optional[np.ndarray] = None,
        is_timestamp_bin_start: bool = False,
    ) -> np.ndarray:
    r"""
        Bin spikes given by an nwb units dataframe.
        There is one bin per input timestamp. If timestamps are not provided, they are inferred from the spike times.
        Timestamps are ideally provided spaced bin_size_s apart.
        If not:
        - if consecutive interval is greater than bin_size_s, then only the proximal bin_size_s interval is used.
        - if consecutive interval is less than bin_size_s, spikes will be binned in the provided interval (i.e. those bins will be smaller than bin_size_s apart).
            - Not outputting repeated spikes mainly because the implementation would be more complex (no single call to np.histogram)
        Args:
            units: df with only index (spike index) and spike times (list of times in seconds). From nwb.units.
            bin_size_s: size of each bin to output in seconds.
            bin_timestamps: array of timestamps indicating bin time in seconds.
            is_timestamp_bin_start: if True, the bin is considered to start at the timestamp, otherwise it ends at the timestamp.
        Returns:
            array of spike counts per bin, per unit. Shape is (bins x units).
    """
    # Make even bins
    if bin_timestamps is None:
        end_time = units.spike_times.apply(lambda s: max(s) if len(s) else 0).max() + bin_size_s
        bin_end_timestamps = np.arange(0, end_time, bin_size_s)
        bin_mask = np.ones(len(bin_end_timestamps), dtype=bool)
    else:
        if is_timestamp_bin_start:
            bin_end_timestamps = bin_timestamps + bin_size_s
        else:
            bin_end_timestamps = bin_timestamps
        # Check contiguous else force cropping for even bins
        gaps = np.diff(bin_end_timestamps)
        if (gaps <= 0).any():
            raise ValueError("bin_end_timestamps must be monotonically increasing.")
        if not np.allclose(gaps, bin_size_s):
            print(f"Warning: Input timestamps not spaced like requested {bin_size_s}. Outputting proximal bin spikes.")
            # Adjust bin_end_timestamps to include bins at the end of discontinuities
            new_bin_ends = [bin_end_timestamps[0]]
            bin_mask = [True] # bool, True if bin ending at this timepoint should be included post mask (not padding)
            for i, gap in enumerate(gaps):
                if not np.isclose(gap, bin_size_s) and gap > bin_size_s:
                    cur_bin_end = bin_end_timestamps[i+1]
                    new_bin_ends.extend([cur_bin_end - bin_size_s, cur_bin_end])
                    bin_mask.extend([False, True])
                else:                        
                    new_bin_ends.append(bin_end_timestamps[i+1])
                    bin_mask.append(True)
            bin_end_timestamps = np.array(new_bin_ends)
            bin_mask = np.array(bin_mask)
        else:
            bin_mask = np.ones(len(bin_end_timestamps), dtype=bool)

    # Make spikes
    spike_arr = np.zeros((bin_mask.sum(), len(units)), dtype=np.uint8)
    bin_edges = np.concatenate([np.array([bin_end_timestamps[0] - bin_size_s]), bin_end_timestamps])
    for idx, (_, unit) in enumerate(units.iterrows()):
        spike_cnt, _ = np.histogram(unit.spike_times, bins=bin_edges)
        if bin_mask is not None:
            spike_cnt = spike_cnt[bin_mask]
        spike_arr[:, idx] = spike_cnt
    return spike_arr


def load_nwb(fn: Union[str, Path], dataset: FalconTask = FalconTask.h1, bin_old=False) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    r"""
        Load data for evaluation.

        Returns:
        - neural_data: binned spike counts. Shape is (time x units)
        - covariates (e.g. kinematics). Shape is (time x n_kinematic_dims)
        - trial_change: boolean of shape (time,) true if the trial has changed
        - eval_mask: boolean array indicating whether to evaluate each time step, True if should
    """
    if not dataset in [FalconTask.h1, FalconTask.h2, FalconTask.m1, FalconTask.m2]:
        raise ValueError(f"Unknown dataset {dataset}")
    with NWBHDF5IO(str(fn), 'r') as io:
        nwbfile = io.read()
        if dataset == FalconTask.h1:
            units = nwbfile.units.to_dataframe()
            kin = nwbfile.acquisition['OpenLoopKinematicsVelocity'].data[:]
            timestamps = nwbfile.acquisition['OpenLoopKinematics'].offset + np.arange(kin.shape[0]) * nwbfile.acquisition['OpenLoopKinematics'].rate
            eval_mask = nwbfile.acquisition['eval_mask'].data[:].astype(bool)
            binned_units = bin_units(units, bin_size_s=0.02, bin_timestamps=timestamps)
            return binned_units, kin, np.zeros(kin.shape[0]), eval_mask
        elif dataset == FalconTask.h2: 
            binned_spikes = nwbfile.acquisition['binned_spikes'].data[()]
            time = nwbfile.acquisition['binned_spikes'].timestamps[()]
            eval_mask = nwbfile.acquisition['eval_mask'].data[()].astype(bool)
            trial_info = (
                nwbfile.trials.to_dataframe()
                .reset_index()
            )
            return binned_spikes, trial_info.cue.values, np.concatenate([[False], np.diff(time) > 0.02]), eval_mask
        elif dataset == FalconTask.m1:
            units = nwbfile.units.to_dataframe()
            raw_emg = nwbfile.acquisition['preprocessed_emg']
            muscles = [ts for ts in raw_emg.time_series]
            emg_data = []
            emg_timestamps = []
            for m in muscles:
                mdata = raw_emg.get_timeseries(m)
                data = mdata.data[:]
                timestamps = mdata.timestamps[:]
                emg_data.append(data)
                emg_timestamps.append(timestamps)
            emg_data = np.vstack(emg_data).T
            emg_timestamps = emg_timestamps[0]
            binned_units = bin_units(units, bin_size_s=0.02, bin_timestamps=emg_timestamps)

            eval_mask = nwbfile.acquisition['eval_mask'].data[:].astype(bool)

            trial_info = (
                nwbfile.trials.to_dataframe()
                .reset_index()
                .rename({"id": "trial_id", "stop_time": "end_time"}, axis=1)
            )
            switch_inds = np.searchsorted(emg_timestamps, trial_info.start_time)
            trial_change = np.zeros(emg_timestamps.shape[0], dtype=bool)
            trial_change[switch_inds] = True

            return binned_units, emg_data, trial_change, eval_mask
        elif dataset == FalconTask.m2:
            units = nwbfile.units.to_dataframe()
            vel_container = nwbfile.acquisition['finger_vel']
            labels = [ts for ts in vel_container.time_series]
            vel_data = []
            vel_timestamps = None
            for ts in labels:
                ts_data = vel_container.get_timeseries(ts)
                vel_data.append(ts_data.data[:])
                vel_timestamps = ts_data.timestamps[:]
            vel_data = np.vstack(vel_data).T
            binned_units = bin_units(units, bin_size_s=0.02, bin_timestamps=vel_timestamps, is_timestamp_bin_start=True)

            eval_mask = nwbfile.acquisition['eval_mask'].data[:].astype(bool)

            trial_change = np.zeros(vel_timestamps.shape[0], dtype=bool)
            trial_info = nwbfile.trials.to_dataframe().reset_index()
            switch_inds = np.searchsorted(vel_timestamps, trial_info.start_time)
            trial_change[switch_inds] = True
            return binned_units, vel_data, vel_timestamps, eval_mask
        else:
            raise NotImplementedError(f"Dataset {dataset} not implemented")
            breakpoint()



#%%
from pathlib import Path
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import numpy as np
from pynwb import NWBHDF5IO
from data_demos.styleguide import set_style
from falcon_challenge.dataloaders import bin_units
set_style()

#%% 
# Load the NWB file
test_path = Path('/snel/share/share/derived/rouse/RTG/NWB_FALCON_v3/held_out_calib')
test_files = sorted(test_path.glob('*.nwb'))

#%% 
def load_nwb(fn: str):
    print(f'Loading {fn}')
    with NWBHDF5IO(fn, 'r') as io:
        nwbfile = io.read()

        trial_info = (
            nwbfile.trials.to_dataframe()
            .reset_index()
            .rename({"id": "trial_id", "stop_time": "end_time"}, axis=1)
        )

        raw_emg = nwbfile.acquisition['preprocessed_emg']
        muscles = [ts for ts in raw_emg.time_series]
        emg_data = []
        emg_timestamps = []
        for m in muscles: 
            mdata = raw_emg.get_timeseries(m)
            data = mdata.data[:]
            timestamps = mdata.timestamps[:]
            emg_data.append(data)
            emg_timestamps.append(timestamps)

        emg_data = np.vstack(emg_data).T 
        emg_timestamps = emg_timestamps[0]

        units = nwbfile.units.to_dataframe()
        binned_units = bin_units(units, bin_size_s=0.02, bin_end_timestamps=emg_timestamps)

        eval_mask = nwbfile.acquisition['eval_mask'].data[:].astype(bool)

        return (
            binned_units,
            emg_data,
            emg_timestamps,
            muscles,
            trial_info,
            eval_mask
        )

#%% 
from sklearn.metrics import r2_score
from decoder_demos.decoding_utils import (
    zscore_data,
    generate_lagged_matrix,
    apply_neural_behavioral_lag,
    fit_and_eval_decoder
)
from preproc.filtering import apply_exponential_filter

#%% 
n_hist = 7

n_trials = [10, 20, 30, 50, 75, 100]

trials = np.full((len(test_files), len(n_trials)+1), np.nan)
perfs = np.full((len(test_files), len(n_trials)+1), np.nan)

for ii in range(len(test_files)): 
    print(f'Processing {test_files[ii]}')
    spikes, emg, time, muscles, tinfo, eval_mask = load_nwb(test_files[ii])

    ds_n_trials = n_trials + [len(tinfo)-1]
    trials[ii, :] = ds_n_trials

    for jj in range(len(n_trials)+ 1): 
        print(f'Decoding {ds_n_trials[jj]} trials')
        ntr = ds_n_trials[jj]
        last_trial = tinfo.iloc[ntr]
        end_time = last_trial['end_time']
        end_ind = np.argmin(np.abs(time - end_time))
        rel_spikes = spikes[:end_ind]
        rel_emg = emg[:end_ind]
        rel_mask = eval_mask[:end_ind]
        rel_ss = apply_exponential_filter(rel_spikes)
        emg_trials = rel_emg[rel_mask]
        ss_trials = rel_ss[rel_mask]
        
        split_ind = int(0.8 * len(emg_trials))
        smoothed_spikes_train, smoothed_spikes_valid = np.split(ss_trials, [split_ind])
        emg_train, emg_valid = np.split(emg_trials, [split_ind])

        train_neural = generate_lagged_matrix(zscore_data(smoothed_spikes_train), n_hist)
        valid_neural = generate_lagged_matrix(zscore_data(smoothed_spikes_valid), n_hist)
        train_behavioral = emg_train[n_hist:]
        valid_behavioral = emg_valid[n_hist:]

        valid_score, _ = fit_and_eval_decoder(
            train_neural,
            train_behavioral,
            valid_neural,
            valid_behavioral,
            return_preds=False
        )

        perfs[ii, jj] = valid_score


#%% 

plt.figure(facecolor='w')
for kk in range(len(test_files)): 
    plt.plot(trials[kk, :], perfs[kk, :], 'o-', label=f'Dataset {kk+1}')
plt.hlines(0.5, 0, 150, color='k', linestyle='--')
plt.ylim(-0.2, 1.2)
plt.xlabel('Number of trials provided')
plt.ylabel('R2 score')
plt.grid(alpha=0.4)

# %%
plt.figure(facecolor='w')
for kk in range(len(test_files)): 
    plt.plot(trials[kk, :] * 0.2, perfs[kk, :], 'o-', label=f'Dataset {kk+1}')
plt.hlines(0.5, 0, 35, color='k', linestyle='--')
plt.ylim(-0.2, 1.2)
plt.xlabel('Number of valid trials for scoring')
plt.ylabel('R2 score')
plt.grid(alpha=0.4)

# %%



#%%

# Realizing now that I need to pull together a separate file for all answers.
# Need a minival.pkl, and a test.pkl
# needs to be a dict of all split targets
from falcon_challenge.config import FalconTask, FalconConfig
from falcon_challenge.dataloaders import load_nwb
from pathlib import Path
import pickle

r"""
    annotations struct:
    {
        'h1': {
            'hash1': {
                'data': np.array, T x K
                'mask': np.array, T
            }
        }
    }
"""

def assemble_phase_answer_key(phase='minival', answer_key_dir='./data/answer_key'):
    annotations = {}
    for dataset in ['h1', 'h2', 'm1', 'm2']:
        annotations[dataset] = {}
        dataset_path = Path(answer_key_dir) / dataset / phase
        dataset_files = list(dataset_path.rglob(f'*{phase}*.nwb'))
        task = getattr(FalconTask, dataset)
        config = FalconConfig(task)
        for d in dataset_files:
            print(d.stem)
            neural_data, decoding_targets, trial_change, eval_mask = load_nwb(d, dataset=task)
            annotations[dataset][config.hash_dataset(d.stem)] = {
                'data': decoding_targets[eval_mask],
                'mask': eval_mask
            }
        print(annotations[dataset].keys())
    return annotations

minival_annotations = assemble_phase_answer_key('minival')
eval_annotations = assemble_phase_answer_key('eval')
# save these as pickles

with open('./data/answer_key/minival.pkl', 'wb') as f:
    pickle.dump(minival_annotations, f)

with open('./data/answer_key/eval.pkl', 'wb') as f:
    pickle.dump(eval_annotations, f)


from typing import List
import os
import pickle
from collections import defaultdict
import logging
import numpy as np
from pathlib import Path
from tqdm import tqdm
from sklearn.metrics import r2_score

from falcon_challenge.config import FalconTask, FalconConfig
from falcon_challenge.interface import BCIDecoder
from falcon_challenge.dataloaders import load_nwb

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

DATASET_HELDINOUT_MAP = {
    'h1': {
        'held_in': [
            'S0_set_1', 
            'S0_set_2', 
            'S1_set_1', 
            'S1_set_2', 
            'S1_set_3',
            'S2_set_1',
            'S2_set_2',
            'S3_set_1',
            'S3_set_2',
            'S4_set_1',
            'S4_set_2',
            'S5_set_1',
            'S5_set_2',
        ],
        'held_out': [
            'S6_set_1', 
            'S6_set_2', 
            'S7_set_1', 
            'S7_set_2', 
            'S8_set_1', 
            'S8_set_2', 
            'S9_set_1', 
            'S9_set_2', 
            'S10_set_1', 
            'S10_set_2', 
            'S11_set_1', 
            'S11_set_2', 
            'S12_set_1',
            'S12_set_2',
        ],
    },
    'm1': {
        'held_in': ['20120924', '20120926', '20120927', '20120928'],
        'held_out': ['20121004', '20121017', '20121022', '20121024'],
    },
    'h2': {

    },
    'm2': {
    },
}

HELD_IN_KEYS = {
    FalconTask.h1: ['S0_', 'S1_', 'S2_', 'S3_', 'S4_', 'S5_'],
    FalconTask.m1: ['L_20120924', 'L_20120926', 'L_20120927', 'L_20120928'],
}

HELD_OUT_KEYS = {
    FalconTask.h1: ['S6_', 'S7_', 'S8_', 'S9_', 'S10_', 'S11_', 'S12_'],
    FalconTask.m1: ['L_20121004', 'L_20121017', 'L_20121022', 'L_20121024'],
}

# Development time flag. False allows direct evaluation without payload writing, only usable for local minival.
# Should be set to true for remote evaluation.
# USE_PKLS = False
USE_PKLS = True

HELDIN_OR_OUT_MAP = {
    'held_in': "Held In",
    'held_out': "Held Out",
}

def evaluate(
    test_annotation_file: str, # The annotation file for the phase
    user_submission_file: str, # * JY: This appears to always be /submission/submission.csv on EvalAI. No matter - load it as a pickle.
    phase_codename: str, # e.g. minival or test
    **kwargs
):
    r"""
        Test struct:
        {
            'h1': {
                'hash': {
                    'data': tgt,
                    'mask': mask,
                }
            }
        }
        User submission struct:
        {
            'h1': {
                'hash': pred,

                'normalized_latency': 1,
            }
        }
        Evaluate payloads with potentially multiple splits worth of data
        - Low pri: can I provide all results or just one split's worth entry? Currently providing 1, examples just provide 1, but in general would be nice to provide all. User shouldn't be able to submit more than 1, though.
    """
    # ! Want: Locally, test_annotation should be somewhere safe (tmp)
    # ! Remotely, it shoudl be /submission/submission.csv exactly.
    # Ignore explicit annotations provided and directly search for concatenated answers
    logger.info(f"Evaluation: Docker side")
    # test_annotation_file = '/dataset/evaluation_data/answer_key/minival.pkl'
    logger.info(f"Loading GT from {test_annotation_file}")
    logger.info(f"Loading submission from {user_submission_file}")
    logger.info(f"Phase: {phase_codename}")

    result = []
    # Load pickles
    try:
        with open(test_annotation_file, 'rb') as test_annotation_file, open(user_submission_file, 'rb') as user_submission_file:
            test_annotations = pickle.load(test_annotation_file)
            user_submission = pickle.load(user_submission_file)
    except Exception as e:
        logger.error(f"Checking root: {os.listdir('/')}")
        raise ValueError(f"Failed to load submission pickles: {e}. dir is {os.getcwd()}; contents {os.listdir()}.")
    for datasplit in user_submission: # datasplit e.g. h1, m1
        if datasplit not in test_annotations:
            raise ValueError(f"Missing {datasplit} in GT labels.")
        split_annotations = test_annotations[datasplit]
        split_result = {}
        split_result["Normalized Latency"] = user_submission[datasplit]["normalized_latency"]
        del user_submission[datasplit]["normalized_latency"]
        pred_dict = defaultdict(list)
        tgt_dict = defaultdict(list)
        mask_dict = defaultdict(list)
        for dataset in user_submission[datasplit]:
            dataset_pred = user_submission[datasplit][dataset]
            dataset_tgt = split_annotations[dataset]['data']
            dataset_mask = split_annotations[dataset]['mask']
            if dataset in DATASET_HELDINOUT_MAP[datasplit]['held_in']:
                pred_dict['held_in'].append(dataset_pred)
                tgt_dict['held_in'].append(dataset_tgt)
                mask_dict['held_in'].append(dataset_mask)
            elif dataset in DATASET_HELDINOUT_MAP[datasplit]['held_out']:
                pred_dict['held_out'].append(dataset_pred)
                tgt_dict['held_out'].append(dataset_tgt)
                mask_dict['held_out'].append(dataset_mask)
            else:
                raise ValueError(f"Dataset {dataset} submitted but not found in held-in or held-out list of split {datasplit}.")
        for in_or_out in pred_dict:
            if len(pred_dict[in_or_out]) < len(DATASET_HELDINOUT_MAP[datasplit][in_or_out]):
                raise ValueError(f"Missing predictions for {datasplit} {in_or_out}. User submitted: {user_submission[datasplit].keys()}. Expecting more like: {HELDIN_OR_OUT_MAP[datasplit][in_or_out]}.")
            pred = np.concatenate(pred_dict[in_or_out])
            tgt = np.concatenate(tgt_dict[in_or_out])
            mask = np.concatenate(mask_dict[in_or_out])
            eval_fn = FalconEvaluator.compute_metrics_classification if 'h2' in datasplit else FalconEvaluator.compute_metrics_regression
            try:
                metrics = eval_fn(pred, tgt, mask)
            except Exception as e:
                raise ValueError(f"Failed to compute metrics for {datasplit} {in_or_out}: {e}")
            for k in metrics:
                split_result[f'{HELDIN_OR_OUT_MAP[in_or_out]} {k}'] = metrics[k]
        result.append({f'{phase_codename}_split_{datasplit}': split_result})

    print(f"Returning result from phase: {phase_codename}: {result}")
    # Out struct according to https://evalai.readthedocs.io/en/latest/evaluation_scripts.html
    return {"result": result, 'submission_result': result[0]}

class FalconEvaluator:

    def __init__(self, eval_remote=False, split='h1'):
        self.eval_remote = eval_remote
        assert split in ['h1', 'h2', 'm1', 'm2'], "Split must be h1, h2, m1, or m2."
        self.dataset: FalconTask = getattr(FalconTask, split)
        self.cfg = FalconConfig(self.dataset)

    @staticmethod
    def get_eval_handles(is_remote: bool, dataset: FalconTask, phase: str = 'minival'):
        if is_remote: # i.e. definitely docker
            data_dir = os.environ.get("EVAL_DATA_PATH")
        else: # possibly docker or local
            if os.path.exists(f"./data/{dataset.name}"):
                logger.info("Using local data directory.")
                data_dir = "data"
            else:
                data_dir = os.environ.get("EVAL_DATA_PATH") # a local docker eval
        data_dir = Path(data_dir) / dataset.name
        if phase == 'test': # TODO wire wherever test is actually stored on remote
            eval_dir = data_dir / f"eval"
        else:
            eval_dir = data_dir / "minival"
        return sorted(list(eval_dir.glob("*val*.nwb")))

    def get_eval_files(self, phase: str = 'minival'):
        logger.info("Searching for evaluation data.")
        handles = self.get_eval_handles(self.eval_remote, self.dataset, phase=phase)
        logger.info(f"Found {len(handles)} files.")
        if len(handles) == 0:
            raise FileNotFoundError(f"No files found in {self.dataset.name} for phase {phase}. Note test phase data is only available on EvalAI remote.")
        return handles
    
    def predict_files(self, decoder: BCIDecoder, eval_files: List):
        # returns triple dict, keyed by datafile hash and contains preds, targets, and eval_mask respective
        # TODO this does not return uniquely identifiable data if eval_files is partial, e.g. if we only has set 2 of a day with 2 sets, we'll happily just provide partial predictions.
        all_preds = defaultdict(list)
        all_targets = defaultdict(list)
        all_eval_mask = defaultdict(list)

        for datafile in tqdm(sorted(eval_files)):
            if not datafile.exists():
                raise FileNotFoundError(f"File {datafile} not found.")
            neural_data, decoding_targets, trial_change, eval_mask = load_nwb(datafile, dataset=self.dataset)
            decoder.reset(dataset=datafile)
            trial_preds = []
            for neural_observations, trial_delta_obs, step_mask in zip(neural_data, trial_change, eval_mask):
                if trial_delta_obs:
                    decoder.on_trial_end()
                if step_mask:
                    trial_preds.append(decoder.predict(neural_observations))
                else:
                    decoder.observe(neural_observations)
                    trial_preds.append(np.full(self.cfg.out_dim, np.nan))
            all_preds[self.cfg.hash_dataset(datafile)].append(np.stack(trial_preds))
            all_targets[self.cfg.hash_dataset(datafile)].append(decoding_targets)
            all_eval_mask[self.cfg.hash_dataset(datafile)].append(eval_mask)
        for k in all_preds:
            all_preds[k] = np.concatenate(all_preds[k])
            all_targets[k] = np.concatenate(all_targets[k])
            all_eval_mask[k] = np.concatenate(all_eval_mask[k])
        return all_preds, all_targets, all_eval_mask

    def evaluate_files(self, decoder: BCIDecoder, eval_files: List):
        all_preds, all_targets, all_eval_mask = self.predict_files(decoder, eval_files)
        metrics = self.compute_metrics(all_preds, all_targets, all_eval_mask)
        return metrics

    def evaluate(
            self, 
            decoder: BCIDecoder, 
            phase: str, 
            held_out_only: bool = False, 
            specific_keys: List = []
        ):
        r"""
            Note: Locally, this can produce metrics, but locally and remotely it should also write a submission file 
            that the actual evaluator on remote uses. The evaluation is done separately on remote.

            held_out_only: Only run predictions on held out
            specific_keys: Overrides held_out_only. Only run predictions on datafiles with specific keys.
        """
        assert phase in ['minival', 'test'], "Phase must be minival or test."
        if phase == 'minival' and (held_out_only or specific_keys):
            logger.warning("Ignoring held_out_only and specific_keys for minival phase.")
            held_out_only = False
        
        np.random.seed(0)
        # ! TODO ideally seed other libraries as well...? Is that our responsibility?
        
        eval_files = self.get_eval_files(phase=phase)
        metrics = {}
        prediction_env_var = "PREDICTION_PATH" if self.eval_remote else "PREDICTION_PATH_LOCAL"
        prediction_path = os.environ.get(prediction_env_var, './local_prediction.pkl')
        if not prediction_path:
            raise ValueError("PREDICTION_PATH not set in remote env which expects it. Cannot forward to separate evaluate runscript.")
        gt_path = os.environ.get("GT_PATH", './local_gt.pkl')
        if not gt_path:
            raise ValueError("GT_PATH not set in remote env which expects it. Cannot forward to separate evaluate runscript.")
            
        if phase == 'test':
            eval_files_held_in = [f for f in eval_files if any(k in f.name for k in HELD_IN_KEYS[self.dataset])]
            eval_files_held_out = [f for f in eval_files if any(k in f.name for k in HELD_OUT_KEYS[self.dataset])]
            assert len(eval_files) == len(eval_files_held_in) + len(eval_files_held_out), f"Mismatch in extracted eval #: Eval file state is not consistent with benchmark creation settings. Found {len(eval_files)} files, {len(eval_files_held_in)} held in, {len(eval_files_held_out)} held out."

            if specific_keys:
                raise NotImplementedError("not sure what metrics to compute for specific keys yet.")
            elif held_out_only:
                eval_files_held_in = []

            all_preds, all_targets, all_eval_mask = self.predict_files(decoder, eval_files_held_out)
            if eval_files_held_in:
                all_preds_held_in, all_targets_held_in, all_eval_mask_held_in = self.predict_files(decoder, eval_files_held_in)
                all_preds.update(all_preds_held_in)
                all_targets.update(all_targets_held_in)
                all_eval_mask.update(all_eval_mask_held_in)

        else:
            all_preds, all_targets, all_eval_mask = self.predict_files(decoder, eval_files)
        
        # Indirect remote setup to satisfy EvalAI interface. Save metrics / comparison to file.
        if USE_PKLS:
            inner_pred = {**all_preds}
            inner_tgt_spoof = { # spoof for local mirror of eval ai path, in reality targets are already compiled on eval ai side.
                k: {
                    'data': all_targets[k][all_eval_mask[k]],
                    'mask': all_eval_mask[k],
                } for k in all_targets
            }
            inner_pred['normalized_latency'] = 1 # TODO - CW insert timing code
            pred_payload = {self.dataset.name: inner_pred}
            truth_payload = {self.dataset.name: inner_tgt_spoof}
        else:
            pass
            
        if USE_PKLS:
            Path(prediction_path).parent.mkdir(parents=True, exist_ok=True)
            with open(prediction_path, 'wb') as f:
                pickle.dump(pred_payload, f)
            Path(gt_path).parent.mkdir(parents=True, exist_ok=True)
            with open(gt_path, 'wb') as f:
                pickle.dump(truth_payload, f)
            import time

            if self.eval_remote:
                print("Sleeping before exiting for remote eval - feel free to interrupt for local eval.", flush=True)
                # Gunjan, EvalAI contact says that current static code eval has an issue where the submission dump is only polled by the EvalAI worker comparison script every 5 minutes
                # Sleep so it's definitely available
                time.sleep(300) 
            else:
                return evaluate(
                    test_annotation_file=gt_path,
                    user_submission_file=prediction_path,
                    phase_codename=phase
                )
        else:
            for k, v in metrics.items():
                logger.info("{}: {}".format(k, v))
        
    @staticmethod
    def compute_metrics_regression(preds, targets, eval_mask):
        # assumes targets are already masked
        preds = preds[eval_mask]
        if not targets.shape[0] == preds.shape[0]:
            raise ValueError(f"Targets and predictions have different lengths: {targets.shape[0]} vs {preds.shape[0]}.")
        return {
            "R2": r2_score(targets, preds, multioutput='variance_weighted'),
            "R2 Std.": 0, # TODO Clay
        }

    @staticmethod
    def compute_metrics_classification(preds, targets, eval_mask):
        preds = preds[eval_mask]
        if not targets.shape[0] == preds.shape[0]:
            raise ValueError(f"Targets and predictions have different lengths: {targets.shape[0]} vs {preds.shape[0]}.")
        return {
            "WER": 1-(preds == targets).mean(),
            "WER Std.": 0, # TODO Clay
        }

    def compute_metrics(self, all_preds, all_targets, all_eval_mask=None):
        r"""
            all_preds: array of shape (n_timesteps, k_dim)
            all_targets: array of shape (n_timesteps, k_dim)
            all_eval_mask: array of shape (n_timesteps, k_dim). True if we should evaluate this timestep.
        """
        if self.dataset in [FalconTask.h1, FalconTask.m1, FalconTask.m2]:
            metrics = self.compute_metrics_regression(all_preds, all_targets, all_eval_mask)
        elif self.dataset in [FalconTask.h2]:
            metrics = self.compute_metrics_classification(all_preds, all_targets, all_eval_mask)
        else:
            raise ValueError(f"Unknown dataset {self.dataset}")
        return metrics


r"""
    JY isn't too clear what happens eval-server side on submission but 1 of 2 patterns seem plausible:
    1. Server receives dockerfile and runs it blind.
    - Submitted docker code should call evaluation with config that points to GT data that only lives remotely.
    - This is the pattern that the Habitat challenge uses.
    In order for the image to access the test data, we can mount the data when we run on eval side.
    `docker run -v /path/on/host:/path/in/container your_image_name`
    2. Evaluator
    - Server receives dockerfile, and runs python evaluate.py --code Dockerfile
    And somehow the evaluator will dynamically import the agent code to run it.
"""


import scipy.signal as signal
import numpy as np
import torch
import torch.nn.functional as F

r"""
    H1 filtering
"""
NEURAL_TAU_MS = 240. # exponential filter from H1 Lab
def apply_exponential_filter(
        x, tau=NEURAL_TAU_MS, bin_size=10, extent: int=1
    ):
    """
    Apply a **causal** exponential filter to the neural signal.

    :param x: NumPy array of shape (time, channels)
    :param tau: Decay rate (time constant) of the exponential filter
    :param bin_size: Bin size in ms (default is 10ms)
    :return: Filtered signal
    :param extent: Number of time constants to extend the filter kernel

    Implementation notes:
    # extent should be 3 for reporting parity, but reference hardcodes a kernel that's equivalent to extent=1
    """
    t = np.arange(0, extent * tau, bin_size)
    # Exponential filter kernel
    kernel = np.exp(-t / tau)
    kernel /= np.sum(kernel)
    # Apply the filter
    filtered_signal = np.array([signal.convolve(x[:, ch], kernel, mode='full')[:len(x)] for ch in range(x.shape[1])]).T
    return filtered_signal


def gaussian_kernel(size, sigma):
    """
    Create a 1D Gaussian kernel.
    """
    size = int(size)
    x = np.linspace(-size // 2, size // 2, size)
    kernel = np.exp(-0.5 * (x / sigma) ** 2)
    kernel /= kernel.sum()
    return kernel

def smooth(position, kernel_size: int, sigma: float):
    """
    Apply Gaussian smoothing on the position data (dim 0)
    kernel_size: size of the kernel
    sigma: standard deviation of the Gaussian kernel
    """
    kernel = gaussian_kernel(kernel_size, sigma)
    pad_total = kernel_size - 1
    pad_left = pad_total // 2
    pad_right = pad_total - pad_left

    position = torch.as_tensor(position, dtype=torch.float32)
    position = F.pad(position.T, (pad_left, pad_right), 'replicate')
    smoothed = F.conv1d(position.unsqueeze(1), torch.as_tensor(kernel, dtype=torch.float32).unsqueeze(0).unsqueeze(0), padding=0)
    return smoothed.squeeze().T.numpy()


# -- frequency filtering functions
def _filter(b, a, x):
    """zero-phase digital filtering that handles nans"""
    nan_mask= np.isnan(x)
    is_nans = (nan_mask.sum() > 0)
    # temporarily replace nans with mean for filtering
    if is_nans:
        x[nan_mask] = np.nanmean(x)

    # apply filtering
    x = signal.filtfilt(b, a, x, axis=0)

    # put nans back where they were
    if is_nans:
        x[nan_mask] = np.nan

    return x

# notch filter
def apply_notch_filt(x, fs, notch_cent_freq, notch_bw_freq):
    """apply recursive notch filtering at specified frequencies"""
    for n_freq, n_bw in zip(notch_cent_freq, notch_bw_freq):
        # python iirnotch expects quality factor (Q)
        Q = n_freq/n_bw
        b, a = signal.iirnotch(n_freq, Q, fs=fs)
        x = _filter(b, a, x)
    return x

# high-pass filter
def apply_butter_filt(x, fs, filt_type, cutoff_freq, filt_order=4):
    """apply 4th order butterworth filtering at specified frequency"""
    # design filter
    b, a = signal.butter(filt_order, cutoff_freq, filt_type, fs=fs)

    # apply filter
    x = _filter(b,a,x)

    return x

# define savistky golay differentiation function
def apply_savgol_diff(x, window_length, polyorder, deriv, delta):
    """apply savitsky-golay differentiation"""

    y = signal.savgol_filter(x, window_length=window_length,
                             polyorder=polyorder,
                             deriv=deriv,
                             delta=delta,
                             mode='constant')
    return y

def apply_clipping(x, clip_quantile):
    """apply clipping to input signal"""
    clip_val = np.quantile(np.abs(x), clip_quantile)
    x = np.clip(x, -clip_val, clip_val)
    return x

def apply_scaling(x, scale_quantile):
    """apply scaling to input signal"""
    scale_val = np.quantile(np.abs(x), scale_quantile)
    x = x/scale_val
    return x

def apply_bayes_filter(x, fs, nbins, sigmax, alpha, beta, obs_model_type, pointmax):
    """runs bayesian filtering on emg data arrays

    Parameters
    ----------
    fs : float
        Sampling frequency (Hz)
    nbins : int
        number of bins to create histogram
    sigmax : float
        maximum value of histogram
    alpha : float
        controls the time-varying diffusion process underlying the latent variable evol.
    beta : float
        controls the probability of larger jumps of the latent variable
    obs_model_type : str
        observation model type. Either 'Gauss' or 'Laplace' currently supported
    pointmax : bool
        whether to return maximum aposterior (MAP) est. if True, else return expectation value

    Returns
    -------
    np.array of bayesian filtered output

    Example
    -------

    """

    def time_evolve_prior(prior, dt, dsigma, alpha, beta):
        """Chapman Kolmogorov stochastic differential equation
        captures dynamics of EMG signal.
        """

        shifted_prior_1 = np.insert(prior[0:-1], 0, prior[0])
        shifted_prior_2 = np.append(prior[1:], prior[-1])
        augmented_prior = (shifted_prior_1 + (-2*prior) + shifted_prior_2)/np.power(dsigma,2)
        alpha_term = dt*alpha*augmented_prior
        beta_term = dt*beta + (1 - dt*beta)* prior
        timeevol_prior = alpha_term + beta_term

        return timeevol_prior

    def compute_likelihood(data, bins, obs_model_type):
        """compute likelihood from respective observation model given data."""

        def gauss_likelihood(data, bins):
            """gaussian likelihood"""
            numerator = np.exp(-0.5 *np.divide(np.power(data,2), np.power(bins, 2)))
            likelihood = np.divide(numerator, bins)

            return likelihood

        def laplace_likelihood(data, bins, obs_model_type):
            """laplacian likelihood"""
            numerator = np.exp(np.divide(-np.abs(data), bins))
            likelihood = np.divide(numerator, bins)

        # measurement/observation update
        if obs_model_type == 'Gauss':
            likelihood = gauss_likelihood(data,bins)
        elif obs_model_type == 'Laplace':
            likelihood = laplace_likelihood(data,bins)
        else:
            print('Unknown option for likelihood model')

        return likelihood
    def compute_posterior(likelihood, prior):
        """compute posterior distribution given likelihood and prior"""
        posterior = np.multiply(likelihood, prior)
        # normalize
        posterior = posterior/np.sum(posterior)

        return posterior
    def point_estimation(posterior, bins, pointmax):
        """compute point estimate used as model output"""
        # point estimation
        if pointmax:
            # maps: maximum aposteriori standard deviation
            index = np.argmax(posterior)
            output = bins[index]
        else:
            # expectation value
            output = sum(posterior * bins)

        return output

    # one order of magnitude larger than stddev
    #sigmax = np.std(x)*10

    # create right bin edges
    bins = np.linspace(0, sigmax, nbins+1)
    bins = np.delete(bins,0) # eliminate 0 bin

    # bin size
    dt = 1/fs
    # histogram bin width
    dsigma = bins[1] - bins[0]

    # init prior (uniform dist.)
    prior = np.ones(nbins)*(1/nbins)

    # create array to hold output
    bf_filt_x = np.empty(x.shape)
    samples = x.values.tolist()
    for i, x_i in tqdm(enumerate(samples)):
        # apply time evolution to prior, compute likelihood, posterior, point estimate
        timeevol_prior = time_evolve_prior(prior, dt, dsigma, alpha, beta)
        likelihood = compute_likelihood(x_i, bins, obs_model_type)
        posterior = compute_posterior(likelihood, timeevol_prior)
        bf_filt_x[i] = point_estimation(posterior, bins, pointmax)

        # set prior for next sample with posterior from this sample
        prior = posterior

    return bf_filt_x

# -- resample functions
def resample_column(x, target_fs, source_fs):
    """Apply order 500 Chebychev filter and then downsample"""
    resampled_x = signal.resample_poly(
        x, int(target_fs), int(source_fs))
    return resampled_x

# -- transform functions
# should likely go in a different file but only transform function for now
def rectify(x):
    """apply absolute value rectification"""
    return np.abs(x)

def mean_center(x):
    """ apply mean centering """
    mean_x = np.nanmean(x,axis=0)
    return x - mean_x


import glob
import subprocess
import shlex

""" Run script to make datasets from raw files into NWB format """

base_dir = "/snel/share/share/data/bg2/t5_handwriting/CORP_data/stability_benchmark"
sessions = glob.glob(base_dir + "/*/*.mat")

for sess in sessions: 
    ext = sess.split('/')[-2]
    sess = '.'.join(sess.split('/')[-1].split('.')[1:-1])
    cmd = f"python h2_preproc.py {sess} {ext}"
    subprocess.run(shlex.split(cmd))




r"""
    NDT2 wrapper. Not for running.
"""
from typing import List
from pathlib import Path
import numpy as np
import torch
import pytorch_lightning as pl
from einops import rearrange

from hydra import compose, initialize_config_module

from falcon_challenge.config import FalconConfig
from falcon_challenge.interface import BCIDecoder

from context_general_bci.utils import suppress_default_registry
suppress_default_registry()
from context_general_bci.config import RootConfig, propagate_config, DataKey, MetaKey
from context_general_bci.dataset import DataAttrs, ContextAttrs
from context_general_bci.subjects import SubjectName
from context_general_bci.contexts.context_info import ExperimentalTask
from context_general_bci.model import load_from_checkpoint
from context_general_bci.model_slim import transfer_model

def format_array_name(subject: str):
    return f'FALCON{subject}-M1'

class NDT2Decoder(BCIDecoder):
    r"""
        Load an NDT2 decoder, prepared in:
        https://github.com/joel99/context_general_bci
    """

    def __init__(
            self,
            task_config: FalconConfig,
            model_ckpt_path: str,
            model_cfg_stem: str,
            zscore_path: str,
            dataset_handles: List[str] = []
        ):
        r"""
            Loading NDT2 requires both weights and model config. Weight loading through a checkpoint is standard.
            Model config is typically stored on wandb, but this is not portable enough. Instead, directly reference the model config file.
        """
        self._task_config = task_config
        self.exp_task = getattr(ExperimentalTask, f'falcon_{task_config.task.name}')
        try:
            initialize_config_module(
                config_module="context_general_bci.config",
                job_name="falcon",
                version_base="1.3",
            )
        except:
            print('Hydra Initialize failed, assuming this is not the first decoder.')
        exp_stem, proper_stem = model_cfg_stem.split('/')[:-1], model_cfg_stem.split('/')[-1]
        exp_stem = '/'.join(exp_stem)
        override_path = f"+exp/{exp_stem}={proper_stem}"
        cfg: RootConfig = compose(config_name="config", overrides=[override_path])

        propagate_config(cfg)
        pl.seed_everything(seed=cfg.seed)

        self.subject = getattr(SubjectName, f'falcon_{task_config.task.name}')
        context_idx = {
            MetaKey.array.name: [format_array_name(self.subject)],
            MetaKey.subject.name: [self.subject],
            MetaKey.session.name: sorted([self._task_config.hash_dataset(handle) for handle in dataset_handles]),
            MetaKey.task.name: [self.exp_task],
        }
        data_attrs = DataAttrs.from_config(cfg.dataset, context=ContextAttrs(**context_idx))
        cfg.model.task.decode_normalizer = zscore_path
        model = load_from_checkpoint(model_ckpt_path, cfg=cfg.model, data_attrs=data_attrs)
        model = transfer_model(model, cfg.model, data_attrs)
        self.model = model.to('cuda:0')
        self.model.eval()

        assert task_config.bin_size_ms == cfg.dataset.bin_size_ms, "Bin size mismatch, transform not implemented."
        self.observation_buffer = torch.zeros((cfg.dataset.max_length_ms // task_config.bin_size_ms, task_config.n_channels), dtype=torch.uint8, device='cuda:0')

    def reset(self, dataset: Path = ""):
        dataset_tag = dataset.stem
        self.set_steps = 0
        self.observation_buffer.zero_()
        self.meta_key = torch.tensor([
            self.model.data_attrs.context.session.index(
                self._task_config.hash_dataset(dataset_tag)
            )], device='cuda:0')

    def predict(self, neural_observations: np.ndarray):
        r"""
            neural_observations: array of shape (n_channels), binned spike counts
        """
        self.observe(neural_observations)
        decoder_in = rearrange(self.observation_buffer[-self.set_steps:], 't c -> 1 t c 1')
        out = self.model(decoder_in, self.meta_key) # Remove batch dim
        return out[0].cpu().numpy()
    
    def observe(self, neural_observations: np.ndarray):
        r"""
            neural_observations: array of shape (n_channels), binned spike counts
            - for timestamps where we don't want predictions but neural data may be informative (start of trial)
        """
        self.set_steps += 1
        self.observation_buffer = torch.roll(self.observation_buffer, -1, dims=0)
        self.observation_buffer[-1] = torch.as_tensor(neural_observations, dtype=torch.uint8, device='cuda:0')

    def on_trial_end(self):
        # reset - for some reason m1 benefits from this
        self.set_steps = 0
        self.observation_buffer.zero_()

if __name__ == "__main__":
    print(f"No train/calibration capabilities in {__file__}, use `context_general_bci` codebase.")


#%%

from pynwb import NWBFile
from pynwb import NWBHDF5IO
import numpy as np
import pandas as pd
import scipy.signal as signal

from scipy.io import loadmat
import h5py, glob, logging, sys, os
from datetime import datetime
from dateutil.tz import tzlocal, gettz
import matplotlib.pyplot as plt
from os import path

from pynwb import TimeSeries
from pynwb.file import Subject
from pynwb.misc import Units
from pynwb.behavior import Position
from pynwb.behavior import BehavioralTimeSeries
from pynwb import ProcessingModule
from nwb_create_utils import (
    create_multichannel_timeseries,
    apply_filt_to_multi_timeseries,
)
from filtering import (
    apply_notch_filt,
    apply_butter_filt,
    apply_savgol_diff,
    apply_clipping,
    apply_scaling,
    resample_column,
    rectify,
)

#%%
logger = logging.getLogger()
logger.setLevel(logging.INFO)

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)

SAVE_PATH = "/snel/share/share/derived/rouse/RTG/NWB_FALCON_v5/"
rouse_base_dir = "/snel/share/share/data/rouse/RTG/"
MONKEY = "L"
EXP_DATE = "20120924"

if len(sys.argv) > 2:
    EXP_DATE = sys.argv[1]
    MONKEY = sys.argv[2]

train_dates = ['20120924', '20120926', '20120927', '20120928']
test_dates = ['20121004', '20121017', '20121022', '20121024']

if EXP_DATE in train_dates:
    IS_TEST_DS = False
if EXP_DATE in test_dates:
    IS_TEST_DS = True

# emg file
emg_mat_path = path.join(rouse_base_dir, f"{MONKEY}{EXP_DATE}_AD_Unrect_EMG.mat")
# spikes file
spk_mat_path = glob.glob(path.join(rouse_base_dir, f"{MONKEY}_*_{EXP_DATE}-data.mat"))[0]
# get file string
# file_id = f"{monkey}_{exp_date}"
# load mat data
f_emg = loadmat(emg_mat_path)
f_spk = h5py.File(spk_mat_path, "r")

def convert_datestr_to_datetime(collect_date):
    date_time = datetime.strptime(collect_date, "%Y%m%d").replace(
        tzinfo=gettz("America/Chicago")
    )

    return date_time

#%% extracting trial info

n_conds = f_emg['EMGInfo']['trial_id'][0].shape[0]
exp_event_ixs = np.concatenate(f_emg['EMGInfo']['file_event_sample'][0])
exp_event_times = (exp_event_ixs/1000).round(4)
trial_start_times = exp_event_times[:,0] - 1
trial_end_times = exp_event_times[:,-1] + 1
gocue_times = exp_event_times[:,0]
move_onset_times = exp_event_times[:,1]
contact_times = exp_event_times[:,2]
reward_times = exp_event_times[:,3]

generic_cond_ids = np.concatenate([ [i]*f_emg['EMGInfo']['trial_id'][0][i].shape[0] for i in range(n_conds)]).tolist()
object_ids = np.concatenate([ [f_emg['EMGSettings']['objects'][0][0][i][0]]*f_emg['EMGInfo']['trial_id'][0][i].shape[0] for i in range(n_conds)]).tolist()
obj_id_name_map = { 128: 'Coaxial', 64: 'Perpendicular', 8: 'Sphere', 1: 'Button'}
obj_id_num_map = { 128: 1, 64: 2, 8: 3, 1: 4}
obj_names = [*map(obj_id_name_map.get, object_ids)]
object_ids = [*map(obj_id_num_map.get, object_ids)]
locations = np.concatenate([ [f_emg['EMGSettings']['locations'][0][0][i][0]]*f_emg['EMGInfo']['trial_id'][0][i].shape[0] for i in range(n_conds)]).tolist()

exp_trial_ids = np.concatenate(f_emg['EMGInfo']['trial_id'][0]).squeeze()
trial_order = np.argsort(trial_start_times)

#%% extracting emg data

emg_data = f_emg["EMGRawData"]

def convert_names_to_list(f, name_field):
    names = f[name_field][0][0][0].tolist()
    names = [name[0] for name in names]
    return names

emg_names = convert_names_to_list(f_emg['EMGSettings'], 'ChanNames')
fs_cont = float(f_emg['EMGSettings']['samp_rate'][0][0][0][0])
t_offset = f_emg['EMGSettings']['analog_start_time'][0][0][0][0]

#%%
channels = np.squeeze(f_spk['SpikeSettings']['channels'][()])
all_spike_times = []
for i in range(1, 97):
    ch_spk_times = []
    where_chan = np.where(channels == i)[0]
    for j in where_chan:
        spike_times = f_spk[f_spk['AllSpikeTimes'][j][0]]
        if spike_times.shape[0] == 1:
            spike_times = spike_times[0]
        else:
            spike_times = np.squeeze(spike_times)
        ch_spk_times.append(spike_times)
    if len(ch_spk_times) > 1:
        ch_spk_times = np.concatenate(ch_spk_times)
    elif len(ch_spk_times) == 1:
        ch_spk_times = ch_spk_times[0]
    all_spike_times.append(ch_spk_times)

#%%
# n_units = np.unique(f_spk['SpikeSettings']['channels'][()]).shape[0]
array_group_by_chan = f_spk['SpikeSettings']['array_by_channel'][0]
array_group_by_chan = [chr(array_id) for array_id in array_group_by_chan.tolist()] # this is for each channels that has spikes
array_group_by_elec = []
for i in range(1, 97):
    if i in channels:
        array_group_by_elec.append(
            array_group_by_chan[np.where(channels == i)[0][0]]
        )
    else:
        array_group_by_elec.append(None)

# elec_id_by_chan = f_spk['SpikeSettings']['unique_channel_num'][0]

#%%

def convert_to_NWB(
    fs_cont,
    trial_start_times,
    trial_end_times,
    gocue_times,
    move_onset_times,
    contact_times,
    reward_times,
    generic_cond_ids,
    object_ids,
    object_names,
    locations,
    exp_trial_ids,
    emg_data,
    emg_names,
    t_offset,
    # n_units,
    # elec_id_by_chan,
    spike_times,
    array_group_by_chan,
    spike_time_thresh=[None, None],
    split_label='in_day_oracle' #calibration, eval
):
    file_id = f'{MONKEY}_{EXP_DATE}_{split_label}'
    logger.info("Creating new NWBFile")
    nwbfile = NWBFile(
        session_description="monkey performing reach-to-grasp task",
        identifier=file_id,
        session_start_time=convert_datestr_to_datetime(EXP_DATE),
        experiment_description="reach-to-grasp center out",
        file_create_date=datetime.now(tzlocal()),
        lab="Rouse",
        institution="University of Kansas",
        experimenter="Dr. Adam Rouse",
    )

    subject = Subject(subject_id=f'Monkey{MONKEY}_{split_label}', species='Rhesus macaque', sex='M', age='P7Y')
    nwbfile.subject = subject
    # === NWBFile Step: add trial info
    logger.info("Adding trial info")
    # add trial information
    nwbfile.add_trial_column(name="gocue_time", description="timing of go cue")
    nwbfile.add_trial_column(name="move_onset_time", description="timing of move onset")
    nwbfile.add_trial_column(name="contact_time", description="timing of contact onset")
    nwbfile.add_trial_column(name="reward_time", description="timing of reward")
    nwbfile.add_trial_column(
        name="result", description="outcome of trial (success/failure/abort)"
    )
    nwbfile.add_trial_column(name="number", description="trial number in rouse dataset")
    nwbfile.add_trial_column(name="tgt_loc", description="location of target (angle)")
    nwbfile.add_trial_column(name="tgt_obj", description="object to grasp at target")
    nwbfile.add_trial_column(
        name="obj_id",
        description="integer id for each object",
    )
    nwbfile.add_trial_column(
        name="condition_id", description="generic integer id across all loc/obj conditions"
    )

    result = 'R'
    for ix in range(exp_trial_ids.size):
        # ix = trial_order[i]
        number = exp_trial_ids[ix]
        start_time = pd.to_datetime(trial_start_times[ix], unit='s').round('20ms').value * 1e-9
        end_time = pd.to_datetime(trial_end_times[ix], unit='s').round('20ms').value * 1e-9
        gocue_time = pd.to_datetime(gocue_times[ix], unit='s').round('20ms').value * 1e-9
        move_onset_time = pd.to_datetime(move_onset_times[ix], unit='s').round('20ms').value * 1e-9
        contact_time = pd.to_datetime(contact_times[ix], unit='s').round('20ms').value * 1e-9
        reward_time = pd.to_datetime(reward_times[ix], unit='s').round('20ms').value * 1e-9
        tgt_loc = locations[ix]
        tgt_obj = obj_names[ix]
        obj_id = object_ids[ix]
        condition_id = generic_cond_ids[ix]
        nwbfile.add_trial(
            start_time=start_time,
            stop_time=end_time,
            gocue_time=gocue_time,
            move_onset_time=move_onset_time,
            contact_time=contact_time,
            reward_time=reward_time,
            result=result,
            number=number,
            tgt_loc=tgt_loc,
            tgt_obj=tgt_obj,
            obj_id=obj_id,
            condition_id=condition_id,
        )

    logger.info("Adding acquisition data")

    target_fs = 50
    t_cont = (np.arange(0, emg_data.shape[0])/fs_cont).round(4)

    # emg_mts = create_multichannel_timeseries(
    #     "emg_raw", emg_names, emg_data, timestamps=t_cont, unit="mV"
    # )
    # nwbfile.add_acquisition(emg_mts)
    # # create processing module
    # emg_filt = nwbfile.create_processing_module(
    #     "emg_filtering_module",
    #     "module to perform emg pre-processing from raw to rectified emg",
    # )

    # extract acquisition data
    # raw_emg = nwbfile.acquisition["emg_raw"]
    raw_emg = emg_data

    notch_cent_freq = [60, 180, 200, 300, 400]
    notch_bw_freq = [2, 2, 2, 2, 2]

    # high-pass filtering
    hp_cutoff_freq = 65  # Hz

    # 1) notch filter
    # notch_emg = apply_filt_to_multi_timeseries(
    #     raw_emg,
    #     apply_notch_filt,
    #     "emg_notch",
    #     fs_cont,
    #     notch_cent_freq,
    #     notch_bw_freq,
    # )
    notch_emg = apply_notch_filt(raw_emg, fs_cont, notch_cent_freq, notch_bw_freq)


    # 2) high pass filter
    # hp_emg = apply_filt_to_multi_timeseries(
    #     notch_emg, apply_butter_filt, "emg_hp", fs_cont, "high", hp_cutoff_freq
    # )
    hp_emg = apply_butter_filt(notch_emg, fs_cont, "high", hp_cutoff_freq)

    # 3) rectify
    # rect_emg = apply_filt_to_multi_timeseries(hp_emg, rectify, "emg")
    rect_emg = rectify(hp_emg)

    # clipping and quantile normalization of EMG
    CLIP_Q = 0.99
    SCALE_Q = 0.95
    clip_emg = apply_clipping(rect_emg, CLIP_Q)
    # clip_emg = apply_filt_to_multi_timeseries(
    #     rect_emg, apply_clipping, 'emg_clip', CLIP_Q
    # )
    # scale_emg = apply_filt_to_multi_timeseries(
    #     clip_emg, apply_scaling, 'emg_scale', SCALE_Q
    # )
    scale_emg = apply_scaling(clip_emg, SCALE_Q)

    # resample all the processed EMG data to 20ms bins (50Hz)
    # resamp_emg = apply_filt_to_multi_timeseries(
    #     scale_emg, resample_column, 'emg_resample', target_fs, fs_cont
    # )
    resamp_emg = resample_column(scale_emg, target_fs, fs_cont)

    # rectify again
    # rerect_emg = apply_filt_to_multi_timeseries(resamp_emg, rectify, 'emg_rerect')
    rerect_emg = rectify(resamp_emg)

    # apply low pass filter
    # preprocessed_emg = apply_filt_to_multi_timeseries(
    #     rerect_emg, apply_butter_filt, 'preprocessed_emg', 50, 'low', 10
    # )
    preprocessed_emg = apply_butter_filt(rerect_emg, target_fs, 'low', 10)

    # add each step to processing module
    # emg_filt.add_container(notch_emg)
    # emg_filt.add_container(hp_emg)
    # emg_filt.add_container(rect_emg)
    # emg_filt.add_container(clip_emg)
    # emg_filt.add_container(scale_emg)
    # emg_filt.add_container(resamp_emg)
    # emg_filt.add_container(rerect_emg)
    # start_time = spike_time_thresh[0] #if spike_time_thresh[0] else 0) * fs_cont / target_fs
    t_new = np.arange(spike_time_thresh[0], spike_time_thresh[-1], 1/target_fs).round(4)
    emg_proc_mts = create_multichannel_timeseries(
        "preprocessed_emg", emg_names, preprocessed_emg, timestamps=t_new, unit="mV"
    )
    nwbfile.add_acquisition(emg_proc_mts)
    # emg_filt.add_container(preprocessed_emg)

    convert_trial_start_time = pd.to_datetime(trial_start_times, unit='s').round('20ms').values.astype('float64') * 1e-9
    convert_trial_end_time = pd.to_datetime(trial_end_times, unit='s').round('20ms').values.astype('float64') * 1e-9
    eval_mask = np.full(t_new.size, False)
    # now add a mask for getting within-trial periods
    for start, stop in zip(convert_trial_start_time, convert_trial_end_time):
        start_ind = np.searchsorted(t_new, start)
        stop_ind = np.searchsorted(t_new, stop)
        eval_mask[start_ind:stop_ind] = True

    nwbfile.add_acquisition(
        TimeSeries(
            name="eval_mask",
            description="Timesteps to KEEP covariates (for training, eval).",
            timestamps=t_new,
            data=eval_mask,
            unit="bool",
        )
    )

    logger.info("Adding spiking data")

    device = nwbfile.create_device(
        name="floating microelectrode_array",
        description="16-electrode array",
        manufacturer="MicroProbes for Life Sciences",
    )
    E_elec_group = nwbfile.create_electrode_group(
        name="E_electrode_group",
        description="Electrodes in an implanted FMA array labeled E",
        location="Motor Cortex",
        device=device,
    )
    F_elec_group = nwbfile.create_electrode_group(
        name="F_electrode_group",
        description="Electrodes in an implanted FMA array labeled F",
        location="Motor Cortex",
        device=device,
    )
    G_elec_group = nwbfile.create_electrode_group(
        name="G_electrode_group",
        description="Electrodes in an implanted FMA array labeled G",
        location="Motor Cortex",
        device=device,
    )
    H_elec_group = nwbfile.create_electrode_group(
        name="H_electrode_group",
        description="Electrodes in an implanted FMA array labeled H",
        location="Motor Cortex",
        device=device,
    )
    I_elec_group = nwbfile.create_electrode_group(
        name="I_electrode_group",
        description="Electrodes in an implanted FMA array labeled I",
        location="Motor Cortex",
        device=device,
    )
    J_elec_group = nwbfile.create_electrode_group(
        name="J_electrode_group",
        description="Electrodes in an implanted FMA array labeled J",
        location="Motor Cortex",
        device=device,
    )
    K_elec_group = nwbfile.create_electrode_group(
        name="K_electrode_group",
        description="Electrodes in an implanted FMA array labeled K",
        location="Motor Cortex",
        device=device,
    )
    L_elec_group = nwbfile.create_electrode_group(
        name="L_electrode_group",
        description="Electrodes in an implanted FMA array labeled L",
        location="Motor Cortex",
        device=device,
    )
    Z_elec_group = nwbfile.create_electrode_group(
        name='Z_electrode_group',
        description="Electrodes not labelled as belonging to any other group",
        location='Motor Cortex',
        device=device,
    )

    elec_group_map = {
        'E': E_elec_group,
        'F': F_elec_group,
        'G': G_elec_group,
        'H': H_elec_group,
        'I': I_elec_group,
        'J': J_elec_group,
        'K': K_elec_group,
        'L': L_elec_group,
        'Z': Z_elec_group,
    }

    nwbfile.units = Units(
        name="units", description="Sampled at 30 kHz with anti-alias", resolution=1/fs_cont
    )
    # orig_elec_ids = np.unique(elec_id_by_chan)
    # new_elec_ids = np.arange(orig_elec_ids.size)
    # elec_id_map = { o_ix: n_ix for (o_ix, n_ix) in zip(orig_elec_ids, new_elec_ids)}
    # array_elec_groups = [*map(elec_group_map.get, array_group_by_chan)]
    array_elec_groups = []
    for arr_id in array_group_by_chan:
        if arr_id is not None:
            array_elec_groups.append(elec_group_map[arr_id])
        else:
            array_elec_groups.append(elec_group_map['Z'])

    for elec_id in range(1, 97):
        # elec_id = orig_elec_ids[i]
        try:
            nwbfile.add_electrode(
                id=elec_id-1, #int(elec_id_map[elec_id]),
                x=np.nan,
                y=np.nan,
                z=np.nan,
                imp=np.nan,
                location="M1",
                group=array_elec_groups[elec_id-1],
                filtering="unknown",
            )
        except ValueError:
            logger.info('Electrode already added to table')

        # where_elec = np.where(elec_id_by_chan == elec_id)[0]
        # all_spike_times = []
        # for j in where_elec:
        #     spike_times = f_spk[f_spk['AllSpikeTimes'][j][0]]
        #     if spike_times.shape[0] == 1:
        #         spike_times = spike_times[0]
        #     else:
        #         spike_times = np.squeeze(spike_times)
        #     spike_times = (spike_times - t_offset).round(4)
        #     all_spike_times.append(spike_times)

        # all_spike_times = np.concatenate(all_spike_times)
        all_spike_times = np.array(spike_times[elec_id - 1])
        if len(all_spike_times) >= 1:
            all_spike_times = (all_spike_times - t_offset).round(4)

        start_cutoff, stop_cutoff = spike_time_thresh
        if start_cutoff is None:
            start_cutoff = t_new[0]
        if stop_cutoff is None:
            stop_cutoff = t_new[-1]
        # ensure spike times are within bound of continuous data
        keep_mask = (all_spike_times > start_cutoff) & (all_spike_times < stop_cutoff)
        valid_spike_times = all_spike_times[keep_mask] if len(all_spike_times) >= 1 else all_spike_times
        # if valid_spike_times is not None:
        nwbfile.add_unit(
            id=elec_id-1,
            spike_times=np.squeeze(valid_spike_times),
            electrodes=[elec_id-1],
            obs_intervals=[[start_cutoff, stop_cutoff]],
        )
        # else:
        #     nwbfile.add_unit(
        #         id=i,
        #         electrodes=[elec_id],
        #     )

    nwb_path = path.join(SAVE_PATH, split_label)
    if not path.exists(nwb_path):
        os.makedirs(nwb_path, mode=0o755)
    save_fname = path.join(nwb_path, file_id + ".nwb")
    logger.info(f"Saving NWB file to {save_fname}")
    # write processed file
    with NWBHDF5IO(save_fname, "w") as io:
        io.write(nwbfile)

#%%

trial_start_times = trial_start_times[trial_order]
trial_end_times = trial_end_times[trial_order]
gocue_times = gocue_times[trial_order]
move_onset_times = move_onset_times[trial_order]
contact_times = contact_times[trial_order]
reward_times = reward_times[trial_order]
generic_cond_ids = np.array(generic_cond_ids)[trial_order]
object_ids = np.array(object_ids)[trial_order]
object_names = np.array(obj_names)[trial_order]
locations = np.array(locations)[trial_order]
exp_trial_ids = np.array(exp_trial_ids)[trial_order]

#%%
# FEW_SHOT_CALIBRATION_RATIO = 0.2
EVAL_RATIO = 0.4
n_trials = exp_event_times.shape[0]
# calibration_num = int(np.ceil(n_trials * FEW_SHOT_CALIBRATION_RATIO))
calibration_num = 10
eval_num = int(n_trials * EVAL_RATIO)

#%%
if IS_TEST_DS:
    logger.info("Creating few-shot calibration split")
    # first calibration num = few shot training
    calib_end_ind_emg = int(trial_end_times[:calibration_num][-1] * fs_cont)
    convert_to_NWB(
        fs_cont,
        trial_start_times[:calibration_num],
        trial_end_times[:calibration_num],
        gocue_times[:calibration_num],
        move_onset_times[:calibration_num],
        contact_times[:calibration_num],
        reward_times[:calibration_num],
        generic_cond_ids[:calibration_num],
        object_ids[:calibration_num],
        object_names[:calibration_num],
        locations[:calibration_num],
        exp_trial_ids[:calibration_num],
        emg_data[:calib_end_ind_emg, :],
        emg_names,
        t_offset,
        all_spike_times,
        array_group_by_elec,
        spike_time_thresh=[0, trial_end_times[:calibration_num][-1]], 
        split_label='held_out_calib' #'in_day_oracle' #calibration, eval
    )

    logger.info("Creating evaluation split")
    # last eval num = evaluation set
    eval_start_ind_emg = int(trial_start_times[-eval_num:][0] * fs_cont)
    convert_to_NWB(
        fs_cont,
        trial_start_times[-eval_num:],
        trial_end_times[-eval_num:],
        gocue_times[-eval_num:],
        move_onset_times[-eval_num:],
        contact_times[-eval_num:],
        reward_times[-eval_num:],
        generic_cond_ids[-eval_num:],
        object_ids[-eval_num:],
        object_names[-eval_num:],
        locations[-eval_num:],
        exp_trial_ids[-eval_num:],
        emg_data[eval_start_ind_emg:, :],
        emg_names,
        t_offset,
        all_spike_times,
        array_group_by_elec,
        spike_time_thresh=[trial_start_times[-eval_num:][0], emg_data.shape[0]/fs_cont],
        split_label='held_out_eval'
    )

    logger.info("Creating in-day oracle split")
    # everything that is not eval set = oracle
    convert_to_NWB(
        fs_cont,
        trial_start_times[:-eval_num],
        trial_end_times[:-eval_num],
        gocue_times[:-eval_num],
        move_onset_times[:-eval_num],
        contact_times[:-eval_num],
        reward_times[:-eval_num],
        generic_cond_ids[:-eval_num],
        object_ids[:-eval_num],
        object_names[:-eval_num],
        locations[:-eval_num],
        exp_trial_ids[:-eval_num],
        emg_data[:eval_start_ind_emg, :],
        emg_names,
        t_offset,
        all_spike_times,
        array_group_by_elec,
        spike_time_thresh=[0, trial_start_times[-eval_num:][0]],
        split_label='held_out_oracle'
    )

else:
    logger.info("Creating full training dataset")
    # first 60%
    eval_start_ind_emg = int(trial_start_times[-eval_num:][0] * fs_cont)
    convert_to_NWB(
        fs_cont,
        trial_start_times[:-eval_num],
        trial_end_times[:-eval_num],
        gocue_times[:-eval_num],
        move_onset_times[:-eval_num],
        contact_times[:-eval_num],
        reward_times[:-eval_num],
        generic_cond_ids[:-eval_num],
        object_ids[:-eval_num],
        object_names[:-eval_num],
        locations[:-eval_num],
        exp_trial_ids[:-eval_num],
        emg_data[:eval_start_ind_emg, :],
        emg_names,
        t_offset,
        all_spike_times,
        array_group_by_elec,
        spike_time_thresh=[0, trial_start_times[-eval_num:][0]],
        split_label='held_in_calib'
    )

    logger.info("Creating eval split")
    # and minival dataset which is last EVAL_RATIO of train data
    # eval_start_ind_emg = int(trial_start_times[-eval_num:][0] * fs_cont)
    convert_to_NWB(
        fs_cont,
        trial_start_times[-eval_num:],
        trial_end_times[-eval_num:],
        gocue_times[-eval_num:],
        move_onset_times[-eval_num:],
        contact_times[-eval_num:],
        reward_times[-eval_num:],
        generic_cond_ids[-eval_num:],
        object_ids[-eval_num:],
        object_names[-eval_num:],
        locations[-eval_num:],
        exp_trial_ids[-eval_num:],
        emg_data[eval_start_ind_emg:, :],
        emg_names,
        t_offset,
        all_spike_times,
        array_group_by_elec,
        spike_time_thresh=[trial_start_times[-eval_num:][0], emg_data.shape[0]/fs_cont],
        split_label='held_in_eval'
    )

    logger.info("Creating smoketest data")
    NUM_ST_TRIALS = 2
    st_ind_emg = int(trial_end_times[:NUM_ST_TRIALS][-1] * fs_cont)
    convert_to_NWB(
        fs_cont,
        trial_start_times[:NUM_ST_TRIALS],
        trial_end_times[:NUM_ST_TRIALS],
        gocue_times[:NUM_ST_TRIALS],
        move_onset_times[:NUM_ST_TRIALS],
        contact_times[:NUM_ST_TRIALS],
        reward_times[:NUM_ST_TRIALS],
        generic_cond_ids[:NUM_ST_TRIALS],
        object_ids[:NUM_ST_TRIALS],
        object_names[:NUM_ST_TRIALS],
        locations[:NUM_ST_TRIALS],
        exp_trial_ids[:NUM_ST_TRIALS],
        emg_data[:st_ind_emg, :],
        emg_names,
        t_offset,
        all_spike_times,
        array_group_by_elec,
        spike_time_thresh=[0, trial_end_times[:NUM_ST_TRIALS][-1]],
        split_label='held_in_minival'
    )


# %%



import abc
from pathlib import Path
import numpy as np

class BCIDecoder:

    batch_size: int = 1

    @abc.abstractmethod
    def reset(self, dataset_tag: str = ""):
        pass

    # @staticmethod
    # def get_file_tag(filepath: Path, is_dandi=False):
    #     if filepath.name.endswith('behavior+ecephys') # clearly dandi
    #     # if is_dandi:
    #         # dandi-style sub-MonkeyL-held-in-calib_ses-20120926_behavior+ecephys.nwb  
    #         return filepath.stem.split('_')[-2].split('-')[-1]
    #     else:
    #         # other style: L_20120924_held_in_eval.nwb
    #         pieces = filepath.stem.split('_')
    #         if pieces[-1] in ['minival', 'calib', 'calibration', 'eval', 'full']:
    #             return '_'.join(pieces[:-1])
    #     return filepath.stem

    @abc.abstractmethod
    def predict(self, neural_observations: np.ndarray) -> np.ndarray:
        r"""
            neural_observations: array of shape (n_channels), binned spike counts
        """
        pass

    def observe(self, neural_observations: np.ndarray):
        r"""
            neural_observations: array of shape (n_channels), binned spike counts
            - for timestamps where we don't want predictions but neural data may be informative (start of trial)
        """
        self.predict(neural_observations)

    @abc.abstractmethod
    def on_trial_end(self):
        # Optional hook available in H2 (handwriting) to allow periodic test-time adaptation
        pass

    def set_batch_size(self, batch_size: int):
        self.batch_size = batch_size


r"""
    Load an sklearn decoder.
    To train, for example:
    `python decoder_demos/sklearn_decoder.py --training_dir data/h1/held_in_calib --calibration_dir data/h1/held_out_calib --mode all`
    To evaluate, see `sklearn_sample.py`
"""
from typing import List, Union, Optional
import argparse
import pickle
import numpy as np
from pathlib import Path

from falcon_challenge.config import FalconConfig, FalconTask
from falcon_challenge.dataloaders import load_nwb
from falcon_challenge.interface import BCIDecoder

from filtering import (
    apply_exponential_filter,
    NEURAL_TAU_MS,
)
from decoder_demos.decoding_utils import (
    TRAIN_TEST,
    generate_lagged_matrix,
    fit_and_eval_decoder,
)

def prepare_train_test(
        binned_spikes: np.ndarray,
        targets: np.ndarray,
        blacklist: Optional[np.ndarray]=None,
        history: int=0,
        ):
    signal = apply_exponential_filter(binned_spikes)

    # Remove timepoints where nothing is happening in the kinematics
    still_times = np.all(np.abs(targets) < 0.001, axis=1)
    if blacklist is not None:
        blacklist = still_times | blacklist
    else:
        blacklist = still_times

    train_x, test_x = np.split(signal, [int(TRAIN_TEST[0] * signal.shape[0])])
    train_y, test_y = np.split(targets, [int(TRAIN_TEST[0] * targets.shape[0])])
    train_blacklist, test_blacklist = np.split(blacklist, [int(TRAIN_TEST[0] * blacklist.shape[0])])

    x_mean, x_std = np.nanmean(train_x, axis=0), np.nanstd(train_x, axis=0)
    x_std[x_std == 0] = 1
    y_mean, y_std = np.nanmean(train_y[~train_blacklist], axis=0), np.nanstd(train_y[~train_blacklist], axis=0)
    y_std[y_std == 0] = 1
    train_x = (train_x - x_mean) / x_std
    test_x = (test_x - x_mean) / x_std
    # train_y = (train_y - y_mean) / y_std # don't standardize y if using var weighted r2
    # test_y = (test_y - y_mean) / y_std

    train_blacklist = train_blacklist | np.isnan(train_y).any(axis=1)
    test_blacklist = test_blacklist | np.isnan(test_y).any(axis=1)
    if np.any(train_blacklist):
        print(f"Invalidating {np.sum(train_blacklist)} of {train_blacklist.shape[0]} timepoints in train")
    if np.any(test_blacklist):
        print(f"Invalidating {np.sum(test_blacklist)} of {test_blacklist.shape[0]} timepoints in test")

    if history > 0:
        train_x = generate_lagged_matrix(train_x, history)
        test_x = generate_lagged_matrix(test_x, history)
        train_y = train_y[history:]
        test_y = test_y[history:]
        if blacklist is not None:
            train_blacklist = train_blacklist[history:]
            test_blacklist = test_blacklist[history:]

    # Now, finally, remove by blacklist
    train_x = train_x[~train_blacklist]
    train_y = train_y[~train_blacklist]
    test_x = test_x[~test_blacklist]
    test_y = test_y[~test_blacklist]

    return train_x, train_y, test_x, test_y, x_mean, x_std, y_mean, y_std

class SKLearnDecoder(BCIDecoder):
    r"""
        Load an sklearn decoder. Assumes the dimensionality is correct.
    """
    def __init__(self, task_config: FalconConfig, model_path: str):
        self._task_config = task_config
        with open(model_path, 'rb') as f:
            payload = pickle.load(f)
            assert payload['task'] == task_config.task
            self.clf = payload['decoder']
            self.history = payload['history'] + 1
            MAX_HISTORY = int(NEURAL_TAU_MS / task_config.bin_size_ms) * 5 # bin size ms
            self.x_mean = payload['x_mean']
            self.x_std = payload['x_std']
            self.raw_history_buffer = np.zeros((MAX_HISTORY, task_config.n_channels))
            self.observation_buffer = np.zeros((self.history, task_config.n_channels))

    def reset(self, dataset: Path = ""):
        if isinstance(self.x_mean, dict):
            dataset_tag =  self._task_config.hash_dataset(dataset.stem)
            if dataset_tag not in self.x_mean:
                raise ValueError(f"Dataset tag {dataset_tag} not found in calibration set {self.x_mean.keys()} - did you calibrate on this dataset?")
            self.local_x_mean = self.x_mean[dataset_tag]
            self.local_x_std = self.x_std[dataset_tag]
        else:
            self.local_x_mean = self.x_mean
            self.local_x_std = self.x_std
        if isinstance(self.clf, dict):
            dataset_tag =  self._task_config.hash_dataset(dataset.stem)
            if dataset_tag not in self.clf:
                raise ValueError(f"Dataset tag {dataset_tag} not found decoder set {self.clf.keys()}")
            self.local_clf = self.clf[dataset_tag]
        else:
            self.local_clf = self.clf
        self.raw_history_buffer = np.zeros_like(self.raw_history_buffer)
        self.observation_buffer = np.zeros_like(self.observation_buffer)

    def predict(self, neural_observations: np.ndarray):
        r"""
            neural_observations: array of shape (n_channels), binned spike counts
        """
        self.observe(neural_observations)
        decoder_in = self.observation_buffer[::-1].copy().flatten().reshape(1, -1) # Reverse since this happens to be how the lagged matrix is formatted
        out = self.local_clf.predict(decoder_in)[0]
        return out

    def observe(self, neural_observations: np.ndarray):
        r"""
            neural_observations: array of shape (n_channels), binned spike counts
            - for timestamps where we don't want predictions but neural data may be informative (start of trial)
        """
        self.raw_history_buffer = np.roll(self.raw_history_buffer, -1, axis=0)
        self.raw_history_buffer[-1] = neural_observations
        smth_history = apply_exponential_filter(self.raw_history_buffer, NEURAL_TAU_MS)
        self.observation_buffer = np.roll(self.observation_buffer, -1, axis=0)
        self.observation_buffer[-1] = (smth_history[-1] - self.local_x_mean) / self.local_x_std

def fit_sklearn_decoder(
    datafiles: List[Path],
    calibration_datafiles: List[Path],
    task_config: FalconConfig,
    save_path: Path,
    history = 0,
):
    r"""
        Fit an sklearn (ridge regresssion) decoder.
    """
    (
        all_neural_data,
        all_covariates,
        all_trial_change,
        all_eval_mask
    ) = zip(*[load_nwb(fn, task_config.task) for fn in datafiles])
    all_neural_data = np.concatenate(all_neural_data)
    all_covariates = np.concatenate(all_covariates)
    all_trial_change = np.concatenate(all_trial_change)
    all_eval_mask = np.concatenate(all_eval_mask)
    (
        train_x,
        train_y,
        test_x,
        test_y,
        x_mean,
        x_std,
        y_mean,
        y_std
    ) = prepare_train_test(all_neural_data, all_covariates, ~all_eval_mask, history=history)
    score, decoder = fit_and_eval_decoder(train_x, train_y, test_x, test_y)
    print(f"CV Fit score: {score:.2f}")
    (
        cal_neural_data,
        _,
        _,
        _,
    ) = zip(*[load_nwb(fn, task_config.task) for fn in [*calibration_datafiles, *datafiles]])
    fns = [task_config.hash_dataset(fn.stem) for fn in [*calibration_datafiles, *datafiles]]
    x_means = {}
    x_stds = {}
    for fn, neural_data in zip(fns, cal_neural_data):
        x_means[fn], x_stds[fn] = np.mean(neural_data, axis=0), np.std(neural_data, axis=0)
        if np.any(x_stds[fn] == 0):
            x_stds[fn][x_stds[fn] == 0] = 1
    decoder_obj = {
        'decoder': decoder,
        'task': task_config.task,
        'history': history,
        'x_mean': x_means,
        'x_std': x_stds,
    }
    with open(save_path, 'wb') as f:
        pickle.dump(decoder_obj, f)
    return save_path

def fit_calibration(
    datafiles: List[Path],
    calibration_datafiles: List[Path],
    task_config: FalconConfig,
    save_path: Path
):
    return fit_sklearn_decoder(
        calibration_datafiles,
        calibration_datafiles,
        task_config,
        save_path
    )

def fit_last_session(
    datafiles: List[Path],
    calibration_datafiles: List[Path],
    task_config: FalconConfig,
    save_path: Path,
    history = 0,
):
    day_unique = set([f.stem.split('_')[0] for f in datafiles])
    last_day = sorted(day_unique)[-1]
    fit_datafiles = [d for d in datafiles if last_day in d.stem]
    return fit_sklearn_decoder(
        fit_datafiles,
        calibration_datafiles,
        task_config,
        save_path,
        history = history
    )

def fit_many_decoders(
    datafiles: List[Path],
    calibration_datafiles: List[Path],
    task_config: FalconConfig,
    save_path: Path,
    history = 0,
):
    all_datafiles = [*calibration_datafiles, *datafiles]
    day_unique = set([f.stem.split('_')[0] for f in all_datafiles])
    all_decoders = {}
    x_means = {}
    x_stds = {}
    for day in sorted(day_unique):
        print(f"Training on day {day}")
        fit_datafiles = [d for d in all_datafiles if day == d.stem.split('_')[0]]
        (
            neural_data,
            covariates,
            trial_change,
            eval_mask
        ) = zip(*[load_nwb(fn, task_config.task) for fn in fit_datafiles])
        neural_data = np.concatenate(neural_data)
        covariates = np.concatenate(covariates)
        trial_change = np.concatenate(trial_change)
        eval_mask = np.concatenate(eval_mask)
        (
            train_x,
            train_y,
            test_x,
            test_y,
            x_mean,
            x_std,
            y_mean,
            y_std
        ) = prepare_train_test(neural_data, covariates, ~eval_mask, history=history)
        score, decoder = fit_and_eval_decoder(train_x, train_y, test_x, test_y)
        for f in fit_datafiles:
            fn = task_config.hash_dataset(f.stem)
            all_decoders[fn] = decoder
            x_means[fn], x_stds[fn] = np.mean(neural_data, axis=0), np.std(neural_data, axis=0)
            if np.any(x_stds[fn] == 0):
                x_stds[fn][x_stds[fn] == 0] = 1
        print(f"CV Fit score: {score:.2f}")
    decoder_obj = {
        'decoder': all_decoders,
        'task': task_config.task,
        'history': history,
        'x_mean': x_means,
        'x_std': x_stds,
    }
    with open(save_path, 'wb') as f:
        pickle.dump(decoder_obj, f)
    return decoder_obj

def main(task, training_dir, calibration_dir, history, mode):
    # Your main function logic here
    if mode == 'all':
        fit_fn = fit_sklearn_decoder
    elif mode == 'last':
        fit_fn = fit_last_session
    elif mode == 'per-day':
        fit_fn = fit_many_decoders # saves down a dict
    else:
        raise ValueError(f"Unknown mode: {mode}")
    training_dir = Path(training_dir)
    calibration_dir = Path(calibration_dir)
    task_config = FalconConfig(
        task=FalconTask.__dict__[task],
    )
    save_path = Path(f'local_data/sklearn_{task_config.task}.pkl')
    datafiles = list(training_dir.glob('*calib*.nwb'))
    calibration_datafiles = list(calibration_dir.glob('*calib*.nwb'))
    fit_fn(datafiles, calibration_datafiles, task_config, save_path, history=history)
    print(f"Saved to {save_path}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Train sklearn decoder')

    parser.add_argument('--task', type=str, choices=['h1', 'm1', 'm2'], help='Task for training')
    parser.add_argument('--training_dir', '-t', type=str, help='Root directory for training files')
    parser.add_argument('--calibration_dir', '-c', type=str, help='Root directory for calibration files')
    parser.add_argument('--mode', '-m', type=str, choices=['all', 'last', 'per-day'], help='Mode for training')
    parser.add_argument('--history', type=int, default=0, help='History for decoder')

    args = parser.parse_args()

    main(args.task, args.training_dir, args.calibration_dir, args.history, args.mode)


from setuptools import setup, find_packages

setup(
    name='falcon_challenge',
    version='0.2.8',

    url='https://github.com/snel-repo/stability-benchmark',
    author='Joel Ye',
    author_email='joelye9@gmail.com',

    packages=find_packages(exclude=['data_demos', 'data']),
    install_requires=[
        'numpy',
        'hydra-core',
        'matplotlib',
        'tqdm',
        'scipy',
        'pandas',
        'seaborn',
        'scikit-learn',
        'pynwb',
        'torch',
    ],
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
)


import numpy as np

from falcon_challenge.config import FalconConfig, FalconTask
from falcon_challenge.interface import BCIDecoder

class RandomDecoder(BCIDecoder):
    r"""
        Proposed flow:
            - Determine the task and find the action space (defining a Gym env, use RLLib, TensorDict?)
            - Sample a random action
    """
    def __init__(self, task_config: FalconConfig):
        self._task_config = task_config

    def predict(self, neural_observations: np.ndarray):
        r"""
            neural_observations: array of shape (n_channels), binned spike counts
        """
        return np.random.rand(self._task_config.out_dim)